@@ -1,11 +1,14 @@
 //! Display implementations for types.
 
+use std::borrow::Cow;
 use std::cell::RefCell;
 use std::collections::hash_map::Entry;
 use std::fmt::{self, Display, Formatter, Write};
 use std::rc::Rc;
 
 use ruff_db::display::FormatterJoinExtension;
+use ruff_db::files::FilePath;
+use ruff_db::source::line_index;
 use ruff_python_ast::str::{Quote, TripleQuotes};
 use ruff_python_literal::escape::AsciiEscape;
 use ruff_text_size::{TextRange, TextSize};
@@ -21,7 +24,7 @@ use crate::types::signatures::{CallableSignature, Parameter, Parameters, Signatu
 use crate::types::tuple::TupleSpec;
 use crate::types::visitor::TypeVisitor;
 use crate::types::{
-    BoundTypeVarInstance, CallableType, IntersectionType, KnownBoundMethodType, KnownClass,
+    BoundTypeVarIdentity, CallableType, IntersectionType, KnownBoundMethodType, KnownClass,
     MaterializationKind, Protocol, ProtocolInstanceType, StringLiteralType, SubclassOfInner, Type,
     UnionType, WrapperDescriptorKind, visitor,
 };
@@ -34,7 +37,9 @@ pub struct DisplaySettings<'db> {
     pub multiline: bool,
     /// Class names that should be displayed fully qualified
     /// (e.g., `module.ClassName` instead of just `ClassName`)
-    pub qualified: Rc<FxHashSet<&'db str>>,
+    pub qualified: Rc<FxHashMap<&'db str, QualificationLevel>>,
+    /// Whether long unions and literals are displayed in full
+    pub preserve_full_unions: bool,
 }
 
 impl<'db> DisplaySettings<'db> {
@@ -54,6 +59,22 @@ impl<'db> DisplaySettings<'db> {
         }
     }
 
+    #[must_use]
+    pub fn truncate_long_unions(self) -> Self {
+        Self {
+            preserve_full_unions: false,
+            ..self
+        }
+    }
+
+    #[must_use]
+    pub fn preserve_long_unions(self) -> Self {
+        Self {
+            preserve_full_unions: true,
+            ..self
+        }
+    }
+
     #[must_use]
     pub fn from_possibly_ambiguous_type_pair(
         db: &'db dyn Db,
@@ -70,7 +91,9 @@ impl<'db> DisplaySettings<'db> {
                     .class_names
                     .borrow()
                     .iter()
-                    .filter_map(|(name, ambiguity)| ambiguity.is_ambiguous().then_some(*name))
+                    .filter_map(|(name, ambiguity)| {
+                        Some((*name, QualificationLevel::from_ambiguity_state(ambiguity)?))
+                    })
                     .collect(),
             ),
             ..Self::default()
@@ -78,6 +101,22 @@ impl<'db> DisplaySettings<'db> {
     }
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum QualificationLevel {
+    ModuleName,
+    FileAndLineNumber,
+}
+
+impl QualificationLevel {
+    const fn from_ambiguity_state(state: &AmbiguityState) -> Option<Self> {
+        match state {
+            AmbiguityState::Unambiguous(_) => None,
+            AmbiguityState::RequiresFullyQualifiedName { .. } => Some(Self::ModuleName),
+            AmbiguityState::RequiresFileAndLineNumber => Some(Self::FileAndLineNumber),
+        }
+    }
+}
+
 #[derive(Debug, Default)]
 struct AmbiguousClassCollector<'db> {
     visited_types: RefCell<FxHashSet<Type<'db>>>,
@@ -92,10 +131,32 @@ impl<'db> AmbiguousClassCollector<'db> {
             }
             Entry::Occupied(mut entry) => {
                 let value = entry.get_mut();
-                if let AmbiguityState::Unambiguous(existing) = value
-                    && *existing != class
-                {
-                    *value = AmbiguityState::Ambiguous;
+                match value {
+                    AmbiguityState::Unambiguous(existing) => {
+                        if *existing != class {
+                            let qualified_name_components = class.qualified_name_components(db);
+                            if existing.qualified_name_components(db) == qualified_name_components {
+                                *value = AmbiguityState::RequiresFileAndLineNumber;
+                            } else {
+                                *value = AmbiguityState::RequiresFullyQualifiedName {
+                                    class,
+                                    qualified_name_components,
+                                };
+                            }
+                        }
+                    }
+                    AmbiguityState::RequiresFullyQualifiedName {
+                        class: existing,
+                        qualified_name_components,
+                    } => {
+                        if *existing != class {
+                            let new_components = class.qualified_name_components(db);
+                            if *qualified_name_components == new_components {
+                                *value = AmbiguityState::RequiresFileAndLineNumber;
+                            }
+                        }
+                    }
+                    AmbiguityState::RequiresFileAndLineNumber => {}
                 }
             }
         }
@@ -104,18 +165,18 @@ impl<'db> AmbiguousClassCollector<'db> {
 
 /// Whether or not a class can be unambiguously identified by its *unqualified* name
 /// given the other types that are present in the same context.
-#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+#[derive(Debug, Clone, PartialEq, Eq)]
 enum AmbiguityState<'db> {
     /// The class can be displayed unambiguously using its unqualified name
     Unambiguous(ClassLiteral<'db>),
     /// The class must be displayed using its fully qualified name to avoid ambiguity.
-    Ambiguous,
-}
-
-impl AmbiguityState<'_> {
-    const fn is_ambiguous(self) -> bool {
-        matches!(self, AmbiguityState::Ambiguous)
-    }
+    RequiresFullyQualifiedName {
+        class: ClassLiteral<'db>,
+        qualified_name_components: Vec<String>,
+    },
+    /// Even the class's fully qualified name is not sufficient;
+    /// we must also include the file and line number.
+    RequiresFileAndLineNumber,
 }
 
 impl<'db> super::visitor::TypeVisitor<'db> for AmbiguousClassCollector<'db> {
@@ -214,21 +275,19 @@ impl<'db> ClassLiteral<'db> {
             settings,
         }
     }
-}
 
-struct ClassDisplay<'db> {
-    db: &'db dyn Db,
-    class: ClassLiteral<'db>,
-    settings: DisplaySettings<'db>,
-}
-
-impl ClassDisplay<'_> {
-    fn class_parents(&self) -> Vec<String> {
-        let body_scope = self.class.body_scope(self.db);
-        let file = body_scope.file(self.db);
-        let module_ast = parsed_module(self.db, file).load(self.db);
-        let index = semantic_index(self.db, file);
-        let file_scope_id = body_scope.file_scope_id(self.db);
+    /// Returns the components of the qualified name of this class, excluding this class itself.
+    ///
+    /// For example, calling this method on a class `C` in the module `a.b` would return
+    /// `["a", "b"]`. Calling this method on a class `D` inside the namespace of a method
+    /// `m` inside the namespace of a class `C` in the module `a.b` would return
+    /// `["a", "b", "C", "<locals of function 'm'>"]`.
+    fn qualified_name_components(self, db: &'db dyn Db) -> Vec<String> {
+        let body_scope = self.body_scope(db);
+        let file = body_scope.file(db);
+        let module_ast = parsed_module(db, file).load(db);
+        let index = semantic_index(db, file);
+        let file_scope_id = body_scope.file_scope_id(db);
 
         let mut name_parts = vec![];
 
@@ -254,8 +313,8 @@ impl ClassDisplay<'_> {
             }
         }
 
-        if let Some(module) = file_to_module(self.db, file) {
-            let module_name = module.name(self.db);
+        if let Some(module) = file_to_module(db, file) {
+            let module_name = module.name(db);
             name_parts.push(module_name.as_str().to_string());
         }
 
@@ -264,19 +323,39 @@ impl ClassDisplay<'_> {
     }
 }
 
+struct ClassDisplay<'db> {
+    db: &'db dyn Db,
+    class: ClassLiteral<'db>,
+    settings: DisplaySettings<'db>,
+}
+
 impl Display for ClassDisplay<'_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        if self
-            .settings
-            .qualified
-            .contains(&**self.class.name(self.db))
-        {
-            for parent in self.class_parents() {
+        let qualification_level = self.settings.qualified.get(&**self.class.name(self.db));
+        if qualification_level.is_some() {
+            for parent in self.class.qualified_name_components(self.db) {
                 f.write_str(&parent)?;
                 f.write_char('.')?;
             }
         }
-        f.write_str(self.class.name(self.db))
+        f.write_str(self.class.name(self.db))?;
+        if qualification_level == Some(&QualificationLevel::FileAndLineNumber) {
+            let file = self.class.file(self.db);
+            let path = file.path(self.db);
+            let path = match path {
+                FilePath::System(path) => Cow::Owned(FilePath::System(
+                    path.strip_prefix(self.db.system().current_directory())
+                        .unwrap_or(path)
+                        .to_path_buf(),
+                )),
+                FilePath::Vendored(_) | FilePath::SystemVirtual(_) => Cow::Borrowed(path),
+            };
+            let line_index = line_index(self.db, file);
+            let class_offset = self.class.header_range(self.db).start();
+            let line_number = line_index.line_index(class_offset);
+            write!(f, " @ {path}:{line_number}")?;
+        }
+        Ok(())
     }
 }
 
@@ -481,9 +560,7 @@ impl Display for DisplayRepresentation<'_> {
                     .display_with(self.db, self.settings.clone()),
                 literal_name = enum_literal.name(self.db)
             ),
-            Type::NonInferableTypeVar(bound_typevar) | Type::TypeVar(bound_typevar) => {
-                bound_typevar.display(self.db).fmt(f)
-            }
+            Type::TypeVar(bound_typevar) => bound_typevar.identity(self.db).display(self.db).fmt(f),
             Type::AlwaysTruthy => f.write_str("AlwaysTruthy"),
             Type::AlwaysFalsy => f.write_str("AlwaysFalsy"),
             Type::BoundSuper(bound_super) => {
@@ -492,9 +569,7 @@ impl Display for DisplayRepresentation<'_> {
                     "<super: {pivot}, {owner}>",
                     pivot = Type::from(bound_super.pivot_class(self.db))
                         .display_with(self.db, self.settings.singleline()),
-                    owner = bound_super
-                        .owner(self.db)
-                        .into_type()
+                    owner = Type::from(bound_super.owner(self.db))
                         .display_with(self.db, self.settings.singleline())
                 )
             }
@@ -516,29 +591,37 @@ impl Display for DisplayRepresentation<'_> {
                 .0
                 .display_with(self.db, self.settings.clone())
                 .fmt(f),
-            Type::TypeAlias(alias) => f.write_str(alias.name(self.db)),
+            Type::TypeAlias(alias) => {
+                f.write_str(alias.name(self.db))?;
+                match alias.specialization(self.db) {
+                    None => Ok(()),
+                    Some(specialization) => specialization
+                        .display_short(self.db, TupleSpecialization::No, self.settings.clone())
+                        .fmt(f),
+                }
+            }
         }
     }
 }
 
-impl<'db> BoundTypeVarInstance<'db> {
+impl<'db> BoundTypeVarIdentity<'db> {
     pub(crate) fn display(self, db: &'db dyn Db) -> impl Display {
-        DisplayBoundTypeVarInstance {
-            bound_typevar: self,
+        DisplayBoundTypeVarIdentity {
+            bound_typevar_identity: self,
             db,
         }
     }
 }
 
-struct DisplayBoundTypeVarInstance<'db> {
-    bound_typevar: BoundTypeVarInstance<'db>,
+struct DisplayBoundTypeVarIdentity<'db> {
+    bound_typevar_identity: BoundTypeVarIdentity<'db>,
     db: &'db dyn Db,
 }
 
-impl Display for DisplayBoundTypeVarInstance<'_> {
+impl Display for DisplayBoundTypeVarIdentity<'_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        f.write_str(self.bound_typevar.typevar(self.db).name(self.db))?;
-        if let Some(binding_context) = self.bound_typevar.binding_context(self.db).name(self.db) {
+        f.write_str(self.bound_typevar_identity.identity.name(self.db))?;
+        if let Some(binding_context) = self.bound_typevar_identity.binding_context.name(self.db) {
             write!(f, "@{binding_context}")?;
         }
         Ok(())
@@ -654,7 +737,7 @@ pub(crate) struct DisplayOverloadLiteral<'db> {
 
 impl Display for DisplayOverloadLiteral<'_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        let signature = self.literal.signature(self.db, None);
+        let signature = self.literal.signature(self.db);
         let type_parameters = DisplayOptionalGenericContext {
             generic_context: signature.generic_context.as_ref(),
             db: self.db,
@@ -731,6 +814,10 @@ impl Display for DisplayFunctionType<'_> {
 }
 
 impl<'db> GenericAlias<'db> {
+    pub(crate) fn display(&'db self, db: &'db dyn Db) -> DisplayGenericAlias<'db> {
+        self.display_with(db, DisplaySettings::default())
+    }
+
     pub(crate) fn display_with(
         &'db self,
         db: &'db dyn Db,
@@ -832,7 +919,6 @@ impl Display for DisplayGenericContext<'_> {
         let variables = self.generic_context.variables(self.db);
 
         let non_implicit_variables: Vec<_> = variables
-            .iter()
             .filter(|bound_typevar| !bound_typevar.typevar(self.db).is_self(self.db))
             .collect();
 
@@ -852,6 +938,10 @@ impl Display for DisplayGenericContext<'_> {
 }
 
 impl<'db> Specialization<'db> {
+    pub fn display(&'db self, db: &'db dyn Db) -> DisplaySpecialization<'db> {
+        self.display_short(db, TupleSpecialization::No, DisplaySettings::default())
+    }
+
     /// Renders the specialization as it would appear in a subscript expression, e.g. `[int, str]`.
     pub fn display_short(
         &'db self,
@@ -1242,6 +1332,44 @@ impl Display for DisplayParameter<'_> {
     }
 }
 
+#[derive(Debug, Copy, Clone)]
+struct TruncationPolicy {
+    max: usize,
+    max_when_elided: usize,
+}
+
+impl TruncationPolicy {
+    fn display_limit(self, total: usize, preserve_full: bool) -> usize {
+        if preserve_full {
+            return total;
+        }
+        let limit = if total > self.max {
+            self.max_when_elided
+        } else {
+            self.max
+        };
+        limit.min(total)
+    }
+}
+
+#[derive(Debug)]
+struct DisplayOmitted {
+    count: usize,
+    singular: &'static str,
+    plural: &'static str,
+}
+
+impl Display for DisplayOmitted {
+    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+        let noun = if self.count == 1 {
+            self.singular
+        } else {
+            self.plural
+        };
+        write!(f, "... omitted {} {}", self.count, noun)
+    }
+}
+
 impl<'db> UnionType<'db> {
     fn display_with(
         &'db self,
@@ -1262,6 +1390,11 @@ struct DisplayUnionType<'db> {
     settings: DisplaySettings<'db>,
 }
 
+const UNION_POLICY: TruncationPolicy = TruncationPolicy {
+    max: 5,
+    max_when_elided: 3,
+};
+
 impl Display for DisplayUnionType<'_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         fn is_condensable(ty: Type<'_>) -> bool {
@@ -1283,12 +1416,27 @@ impl Display for DisplayUnionType<'_> {
             .filter(|element| is_condensable(*element))
             .collect::<Vec<_>>();
 
+        let total_entries =
+            usize::from(!condensed_types.is_empty()) + elements.len() - condensed_types.len();
+
+        assert_ne!(total_entries, 0);
+
         let mut join = f.join(" | ");
 
+        let display_limit =
+            UNION_POLICY.display_limit(total_entries, self.settings.preserve_full_unions);
+
         let mut condensed_types = Some(condensed_types);
+        let mut displayed_entries = 0usize;
+
         for element in elements {
+            if displayed_entries >= display_limit {
+                break;
+            }
+
             if is_condensable(*element) {
                 if let Some(condensed_types) = condensed_types.take() {
+                    displayed_entries += 1;
                     join.entry(&DisplayLiteralGroup {
                         literals: condensed_types,
                         db: self.db,
@@ -1296,6 +1444,7 @@ impl Display for DisplayUnionType<'_> {
                     });
                 }
             } else {
+                displayed_entries += 1;
                 join.entry(&DisplayMaybeParenthesizedType {
                     ty: *element,
                     db: self.db,
@@ -1304,6 +1453,17 @@ impl Display for DisplayUnionType<'_> {
             }
         }
 
+        if !self.settings.preserve_full_unions {
+            let omitted_entries = total_entries.saturating_sub(displayed_entries);
+            if omitted_entries > 0 {
+                join.entry(&DisplayOmitted {
+                    count: omitted_entries,
+                    singular: "union element",
+                    plural: "union elements",
+                });
+            }
+        }
+
         join.finish()?;
 
         Ok(())
@@ -1315,23 +1475,45 @@ impl fmt::Debug for DisplayUnionType<'_> {
         Display::fmt(self, f)
     }
 }
-
 struct DisplayLiteralGroup<'db> {
     literals: Vec<Type<'db>>,
     db: &'db dyn Db,
     settings: DisplaySettings<'db>,
 }
 
+const LITERAL_POLICY: TruncationPolicy = TruncationPolicy {
+    max: 7,
+    max_when_elided: 5,
+};
+
 impl Display for DisplayLiteralGroup<'_> {
     fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
         f.write_str("Literal[")?;
-        f.join(", ")
-            .entries(
-                self.literals
-                    .iter()
-                    .map(|ty| ty.representation(self.db, self.settings.singleline())),
-            )
-            .finish()?;
+
+        let total_entries = self.literals.len();
+
+        let display_limit =
+            LITERAL_POLICY.display_limit(total_entries, self.settings.preserve_full_unions);
+
+        let mut join = f.join(", ");
+
+        for lit in self.literals.iter().take(display_limit) {
+            let rep = lit.representation(self.db, self.settings.singleline());
+            join.entry(&rep);
+        }
+
+        if !self.settings.preserve_full_unions {
+            let omitted_entries = total_entries.saturating_sub(display_limit);
+            if omitted_entries > 0 {
+                join.entry(&DisplayOmitted {
+                    count: omitted_entries,
+                    singular: "literal",
+                    plural: "literals",
+                });
+            }
+        }
+
+        join.finish()?;
         f.write_str("]")
     }
 }
@@ -1582,7 +1764,7 @@ mod tests {
 
         let iterator_synthesized = typing_extensions_symbol(&db, "Iterator")
             .place
-            .ignore_possibly_unbound()
+            .ignore_possibly_undefined()
             .unwrap()
             .to_instance(&db)
             .unwrap()
